//
//
//    Copyright (C) 2020 Dave Freese, W1HKJ
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef __WIN32__
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h>
#include <glob.h>
#include <dirent.h>
#endif

#include "flio.h"
#include "flio_ui.h"
#include "serial.h"

using namespace std;

int ACTIVATED;
char input_buffer[2048];
char command[1024];

const int QUEUE_MAX = 1000;
char Message_Queue[QUEUE_MAX];
int In_Ptr = 0;
int Out_Ptr = 0;

Cserial serial;

int serial_fd;

int process_message_Index = 0;

//----------------------------------------------------------------------
void Clear_Serial_Input_Data ( void  )
{
	serial.FlushBuffer();
}

//----------------------------------------------------------------------
void Close_Serial( void )
{
	serial.ClosePort();
}

int Init_Serial( void )
{
	if (progstatus.serial_port == "NONE")
		return 1;

	serial.Device(progstatus.serial_port);

	serial.Baud(atoi(progstatus.serial_baud.c_str()));

	serial.Stopbits(atoi(progstatus.stopbits.c_str()));

	if (progstatus.handshake == "RTSCTS") {
		serial.RTSCTS(true);
		serial.XonXoff(false);
	} else if (progstatus.handshake == "XONXOFF") {
		serial.RTSCTS(false);
		serial.XonXoff(true);
	} else {
		serial.RTSCTS(false);
		serial.XonXoff(false);
	}

	if (!serial.OpenPort()) {
		serial.ClosePort();
		box_arduino_connected->color(FL_BACKGROUND2_COLOR);
		box_arduino_connected->redraw();
		return 1;
	}

	box_arduino_connected->color(FL_GREEN);
	box_arduino_connected->redraw();

	Clear_Serial_Input_Data();
	return 0;
}

//----------------------------------------------------------------------
void Send_Serial ( std::string s )
{
	if (!serial.IsOpen()) return;

	serial.WriteBuffer(s.c_str(), s.length());
	return;
}

//======================================================================
// init_port_combo
//======================================================================

#ifdef __WIN32__
static bool open_serial(const char* dev)
{
	bool ret = false;
	HANDLE fd = CreateFile(dev, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
	if (fd != INVALID_HANDLE_VALUE) {
		CloseHandle(fd);
		ret = true;
	}
	return ret;
}

#  define TTY_MAX 255
void init_port_combos()
{
	char ttyname[21];
	const char tty_fmt[] = "//./COM%u";

	sel_comm_port->clear();
	sel_comm_port->add("NONE");

	for (unsigned j = 0; j < TTY_MAX; j++) {
		snprintf(ttyname, sizeof(ttyname), tty_fmt, j);
		if (!open_serial(ttyname))
			continue;
		snprintf(ttyname, sizeof(ttyname), "COM%u", j);
		sel_comm_port->add(ttyname);
	}
	sel_comm_port->sort();
}
#endif //__WIN32__

//======================================================================
// Linux init_port_combos
//======================================================================

#ifdef __linux__
#ifndef PATH_MAX
#  define PATH_MAX 1024
#endif
#  define TTY_MAX 8

void init_port_combos()
{
	struct stat st;
	char ttyname[PATH_MAX + 1];
	bool ret = false;

	DIR* sys = NULL;
	char cwd[PATH_MAX] = { '.', '\0' };

	sel_comm_port->clear();
	sel_comm_port->add("NONE");

	glob_t gbuf;
	glob("/dev/serial/by-id/*", 0, NULL, &gbuf);
	for (size_t j = 0; j < gbuf.gl_pathc; j++) {
		if ( !(stat(gbuf.gl_pathv[j], &st) == 0 && S_ISCHR(st.st_mode)) ||
		     strstr(gbuf.gl_pathv[j], "modem") )
			continue;
		sel_comm_port->add(gbuf.gl_pathv[j]);
	}
	globfree(&gbuf);

	if (getcwd(cwd, sizeof(cwd)) == NULL) goto out;

	if (chdir("/sys/class/tty") == -1) goto check_cuse;
	if ((sys = opendir(".")) == NULL) goto check_cuse;

	ssize_t len;
	struct dirent* dp;

	while ((dp = readdir(sys))) {
#  ifdef _DIRENT_HAVE_D_TYPE
		if (dp->d_type != DT_LNK)
			continue;
#  endif
		if ((len = readlink(dp->d_name, ttyname, sizeof(ttyname)-1)) == -1)
			continue;
		ttyname[len] = '\0';
		if (!strstr(ttyname, "/devices/virtual/")) {
			snprintf(ttyname, sizeof(ttyname), "/dev/%s", dp->d_name);
			if (stat(ttyname, &st) == -1 || !S_ISCHR(st.st_mode))
				continue;
			sel_comm_port->add(ttyname);
			ret = true;
		}
	}

check_cuse:
	if (sys) {
		closedir(sys);
		sys = NULL;
	}
	if (chdir("/sys/class/cuse") == -1) goto out;
	if ((sys = opendir(".")) == NULL) goto out;

	while ((dp = readdir(sys))) {
#  ifdef _DIRENT_HAVE_D_TYPE
		if (dp->d_type != DT_LNK)
			continue;
#  endif
		if ((len = readlink(dp->d_name, ttyname, sizeof(ttyname)-1)) == -1)
			continue;
		ttyname[len] = '\0';
		if (strstr(ttyname, "/devices/virtual/") && !strncmp(dp->d_name, "mhuxd", 5)) {
			char *name = strdup(dp->d_name);
			if(!name)
				continue;
			char *p = strchr(name, '!');
			if(p)
				*p = '/';
			snprintf(ttyname, sizeof(ttyname), "/dev/%s", name);
			free(name);
			if (stat(ttyname, &st) == -1 || !S_ISCHR(st.st_mode))
				continue;
			sel_comm_port->add(ttyname);
			ret = true;
		}
	}

out:

	std::string tty_virtual = HomeDir;
	tty_virtual.append("vdev");

	tty_virtual.append("/ttyS%u");
	for (unsigned j = 0; j < TTY_MAX; j++) {
		snprintf(ttyname, sizeof(ttyname), tty_virtual.c_str(), j);
		if ( !(stat(ttyname, &st) == 0 && S_ISCHR(st.st_mode)) )
			continue;
		sel_comm_port->add(ttyname);
	}

	if (sys) closedir(sys);

	if ((chdir(cwd) != -1) && !ret) {
		const char* tty_fmt[] = {
			"/dev/ttyS%u",
			"/dev/ttyUSB%u",
			"/dev/usb/ttyUSB%u"
		};
		for (size_t i = 0; i < sizeof(tty_fmt)/sizeof(*tty_fmt); i++) {
			for (unsigned j = 0; j < TTY_MAX; j++) {
				snprintf(ttyname, sizeof(ttyname), tty_fmt[i], j);
				if ( !(stat(ttyname, &st) == 0 && S_ISCHR(st.st_mode)) )
					continue;
				sel_comm_port->add(ttyname);
			}
		}
	}
	sel_comm_port->sort();

}
#endif // __linux__

//======================================================================
// APPLE init_port_combos
//======================================================================

#ifdef __APPLE__
#ifndef PATH_MAX
#  define PATH_MAX 1024
#endif

void init_port_combos()
{
	struct stat st;

	const char* tty_fmt[] = {
		"/dev/cu.*",
		"/dev/tty.*"
	};

	glob_t gbuf;

	sel_comm_port->clear();
	sel_comm_port->add("NONE");

	for (size_t i = 0; i < sizeof(tty_fmt)/sizeof(*tty_fmt); i++) {
		glob(tty_fmt[i], 0, NULL, &gbuf);
		for (size_t j = 0; j < gbuf.gl_pathc; j++) {
			if ( !(stat(gbuf.gl_pathv[j], &st) == 0 && S_ISCHR(st.st_mode)) ||
			     strstr(gbuf.gl_pathv[j], "modem") )
				continue;
//			LOG_WARN("Found serial port %s", gbuf.gl_pathv[j]);
			sel_comm_port->add(gbuf.gl_pathv[j]);
		}
		globfree(&gbuf);
	}
	sel_comm_port->sort();
}
#endif //__APPLE__

//======================================================================
// FreeBSD init_port_combos
//======================================================================

#ifdef __FreeBSD__
#ifndef PATH_MAX
#  define PATH_MAX 1024
#endif
#  define TTY_MAX 8

void init_port_combos()
{
	int retval;
	struct stat st;
	char ttyname[PATH_MAX + 1];
	const char* tty_fmt[] = {
		"/dev/ttyd%u"
	};

	sel_comm_port->clear();
	sel_comm_port->add("NONE");

	for (size_t i = 0; i < sizeof(tty_fmt)/sizeof(*tty_fmt); i++) {
		for (unsigned j = 0; j < TTY_MAX; j++) {
			snprintf(ttyname, sizeof(ttyname), tty_fmt[i], j);
			if ( !(stat(ttyname, &st) == 0 && S_ISCHR(st.st_mode)) )
				continue;
//			LOG_WARN("Found serial port %s", ttyname);
			sel_comm_port->add(ttyname);
		}
	}
	sel_comm_port->sort();
}
#endif //__FreeBSD__


