// ----------------------------------------------------------------------------
// Copyright (C) 2018
//              David Freese, W1HKJ
//
// This file is part of flio.
//
// flio is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// flio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

#include <cstdlib>
#include <string>
#include <vector>
#include <queue>

#include "config.h"
#include "xml_io.h"
#include "flio_status.h"
#include "flio.h"
#include "flio_ui.h"
#include "threads.h"
#include "XmlRpc.h"
#include "util.h"

using XmlRpc::XmlRpcValue;

pthread_mutex_t mutex_xmlrpc = PTHREAD_MUTEX_INITIALIZER;
static pthread_t DIGI_thread;

static const double TIMEOUT = 1.0;

static XmlRpc::XmlRpcClient* client;
static XmlRpcValue* status_query;

#define RIG_UPDATE_INTERVAL  50 // milliseconds
#define CHECK_UPDATE_COUNT   (2000 / RIG_UPDATE_INTERVAL)

bool run_digi_loop = true;
bool fldigi_online = false;

bool ptt_on = false;
bool xmlvfo_changed = false;

int try_count = 1;

void set_fldigi_connect (void *d)
{
	bool b = (long *)d;
	box_fldigi_connected->color(b ? FL_GREEN : FL_BACKGROUND2_COLOR);
	box_fldigi_connected->redraw();
}

//=====================================================================
// socket ops
//=====================================================================

void open_xmlrpc()
{
	int server_port = atoi(progstatus.server_port.c_str());
	client = new XmlRpc::XmlRpcClient(
				progstatus.server_addr.c_str(),
				server_port );
	XmlRpc::setVerbosity(0);
}

void close_xmlrpc()
{
	guard_lock gl_xmlclose(&mutex_xmlrpc);

	delete client;
	client = NULL;
	delete status_query;
	status_query = NULL;
}

static inline void execute(const char* name, const XmlRpcValue& param, XmlRpcValue& result)
{
	if (client) {
		if (!client->execute(name, param, result, TIMEOUT)) {
			fldigi_online = false;
			throw XmlRpc::XmlRpcException(name);
		} else
			fldigi_online = true;
	}
}

// --------------------------------------------------------------------
// receive functions
// --------------------------------------------------------------------

static void get_fldigi_version()
{
	XmlRpcValue result;
	XmlRpcValue param((int)0);
	try {
		execute("fldigi.version", param, result);
		std::string version = std::string(result);
		Fl::awake(set_fldigi_connect, (void *)1);
	} catch (const XmlRpc::XmlRpcException& e) {
		throw e;
	}
}

static std::string buffer;
static void addtext(void *)
{
	update_flio_text(buffer);
}

static void get_fldigi_data_stream()
{
	XmlRpcValue stream;
	XmlRpcValue param((int)0);
	buffer.clear();
	try {
		if (progstatus.data_type == "Rx")
			execute("rx.get_data", param, stream);
		else if (progstatus.data_type == "Tx")
			execute("tx.get_data", param, stream);
		else
			execute("rxtx.get_data", param, stream);

		const XmlRpcValue::BinaryData &data = stream;
		size_t n = data.size();
		if (n) {
			for (size_t i = 0; i < n; i++) {
				if (data[i] == '\r')
					buffer += '\n';
				else if (data[i] >= ' ' && data[i] <= '~')
					buffer += data[i];
			}
			Fl::awake(addtext);
		}
		Fl::awake(set_fldigi_connect, (void *)1);

	} catch (const XmlRpc::XmlRpcException& e) {
		Fl::awake(set_fldigi_connect, (void *)0);
		throw e;
	}
}

static bool exit_loop = false;

void * digi_loop(void *d)
{
	for (;;) {
		if (exit_loop)
			return NULL;

		MilliSleep(50);
		if (!fldigi_online) {
			if (--try_count == 0) {
				try {
					guard_lock gl_xml(&mutex_xmlrpc);
					get_fldigi_version();
				} catch (const XmlRpc::XmlRpcException& e) {
					try_count = CHECK_UPDATE_COUNT;
					Fl::awake(set_fldigi_connect, (void *)0);
				}
			}
		} else {
			try {
				guard_lock gl_xml(&mutex_xmlrpc);
				get_fldigi_data_stream();
			} catch (const XmlRpc::XmlRpcException& e) {
				try_count = CHECK_UPDATE_COUNT;
				Fl::awake(set_fldigi_connect, (void *)0);
			}
		}
	}
	return NULL;
}

void start_digi_loop(void *)
{
	open_xmlrpc();

	if (pthread_create(&DIGI_thread, NULL, digi_loop, NULL) < 0) {
		printf("create thread failed\n");
		return;
	}
}

void stop_digi_loop()
{
	exit_loop = true;
}
